/**
 * SWE KEx-DatexII (de.bsvrz.kex.datex2)
 * 
 * Copyright (C) 2014 TraffGo Road GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * TraffGo Road GmbH
 * Ostwall 11-19
 * 47798 Krefeld
 * Phone: +49-2151-3295058
 * mailto: info@traffgoroad.de
 */

package de.bsvrz.kex.datex2.plugins.konvertierung.impl;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.datex2.allgemein.ObjekteAtgAsp;
import de.bsvrz.kex.datex2.plugins.definition.PluginImport;
import de.bsvrz.kex.datex2.plugins.konvertierung.DaVKonvertierungDynamisch;
import de.bsvrz.sys.funclib.debug.Debug;
import eu.datex2.schema.dynamisch.messquerschnitte.verkehrslage.D2LogicalModel;
import eu.datex2.schema.dynamisch.messquerschnitte.verkehrslage.ElaboratedData;
import eu.datex2.schema.dynamisch.messquerschnitte.verkehrslage.ElaboratedDataPublication;
import eu.datex2.schema.dynamisch.messquerschnitte.verkehrslage.LocationByReference;
import eu.datex2.schema.dynamisch.messquerschnitte.verkehrslage.TrafficStatus;
import eu.datex2.schema.dynamisch.messquerschnitte.verkehrslage.Validity;

/**
 * Klasse implementiert Konvertierung von DATEX-II nach Datenverteiler.
 * 
 * @author TraffGo Road GmbH
 * @version $Id$
 *
 */

public class ImportVerkehrslage implements DaVKonvertierungDynamisch, ClientSenderInterface {
  /** Umrechnung Millisekunden => Sekunden. */
  private static final int FACTOR_MILLI_TO_SECOND = 1000;
  
  /** Umrechnung [0;1] => Prozent [0;100]. */
  private static final float FACTOR_PERCENT = 100.0f;
  
  /** Logging. */
  private static final Debug LOGGER = Debug.getLogger();
  
  /** Zu verwendende Attributgruppe. */
  private static final String ATG_MQ_ZUSTAND = "atg.strfallZustand";
  
  /** Referenz auf Datenverteiler-Verbindung. */
  private ClientDavInterface davConnection;
  
  /** Referenz auf Plugin-Definitions-Klasse. */
  private PluginImport plugin;
  
  /** Zu verwendende Attributgruppe (als Datenverteiler-Objekt). */
  private AttributeGroup atg;
  
  /** Pid des zu verwendenden Aspekts. */
  private String aspPid;
  
  /** Prfix fr Log-Ausgaben. */
  private String logPrefix = "";
  
  /**
   * Konstruktor.
   * 
   * @param plugin Referenz auf Plugin-Definitions-Klasse
   */
  
  public ImportVerkehrslage(final PluginImport plugin) {
    this.plugin = plugin;
  }

  @Override
  public void liefereDaten(final Object xml) {
    int sendTotal = 0;
    try {
      D2LogicalModel d2l;
      
      if (!(xml instanceof JAXBElement<?>)) {
        try {
          d2l = (D2LogicalModel) xml;
        } catch (Exception ex) {
          throw new Exception("Fehlerhafte XML-Struktur bergeben.", ex);
        }
      } else {
        JAXBElement<?> root = (JAXBElement<?>) xml;

        if (!(root.getValue() instanceof D2LogicalModel)) {
          throw new Exception("Fehlerhafte XML-Struktur bergeben.");        
        }
        d2l = (D2LogicalModel) root.getValue();
      }
      
      ElaboratedDataPublication payload = (ElaboratedDataPublication) d2l.getPayloadPublication();

      LOGGER.info(logPrefix + "Importiere Daten (" + payload.getElaboratedDatas().size() + " Datenstze)");

      for (ElaboratedData elaboratedData : payload.getElaboratedDatas()) {
        try {
          TrafficStatus trafficStatus = (TrafficStatus) elaboratedData.getBasicData();
          LocationByReference locationByReference = (LocationByReference) trafficStatus.getPertinentLocation();
          String pid = locationByReference.getPredefinedLocationReference().getId();
          String situation = null;
          Float accuracy = trafficStatus.getTrafficStatus().getAccuracy();
          if (accuracy != null) {
            accuracy /= FACTOR_PERCENT;
          } 

          switch (trafficStatus.getTrafficStatus().getTrafficStatusValue()) {
          case CONGESTED : situation = "stockender Verkehr"; break;
          case FREE_FLOW : situation = "freier Verkehr"; break;
          case HEAVY : situation = "lebhafter Verkehr"; break;
          case UNKNOWN : situation = "keine Aussage"; break;
          case IMPOSSIBLE : situation = "Stau"; break;
          default: throw new Exception("Wert unbekannt");
          }

          Validity valid = elaboratedData.getValidity();

          int horizont = 0;
          if ((elaboratedData.isForecast() != null) && (elaboratedData.isForecast())) {
            horizont = 
                valid.getValidityTimeSpecification().getOverallEndTime().getMillisecond()
                - valid.getValidityTimeSpecification().getOverallStartTime().getMillisecond();
          }

          Float periodInSeconds = trafficStatus.getMeasurementOrCalculationPeriod();
          if (periodInSeconds == null) {
            periodInSeconds = 0f;
          }
          int periodInMillis = (int) (periodInSeconds * FACTOR_MILLI_TO_SECOND);

          Data d2 = davConnection.createData(atg);

          d2.getItem("Situation").asTextValue().setText(situation);
          d2.getItem("T").asTimeValue().setMillis(periodInMillis);
          d2.getItem("Horizont").asTimeValue().setMillis(horizont);
          if (accuracy == null) {
            d2.getItem("Gte").getItem("Index").asTextValue().setText("nicht ermittelbar");
          } else {
            d2.getItem("Gte").getItem("Index").asScaledValue().set(accuracy);          
          }
          d2.getItem("Gte").getItem("Verfahren").asTextValue().setText("Standard");

          sendData(pid, ATG_MQ_ZUSTAND, aspPid, d2);
          sendTotal++;
        } catch (Exception e) {
          LOGGER.warning(logPrefix + "Unerwarteter Fehler beim Import eines einzelnen Datensatzes: ", e);
        } 
      }
      LOGGER.info(logPrefix + "Import abgeschlossen (" + sendTotal + " Datenstze importiert)");
    } catch (Exception ex) {
      LOGGER.error(logPrefix + "Allgemeiner Fehler beim Import", ex);
      plugin.getErweiterterLogger().sendeAnOperator(logPrefix 
          + "Datenimport vom MDM konnte nicht durchgefhrt werden " + ex.getMessage());
    }
  }
  
  /**
   * Sende Daten an den Datenverteiler.
   * 
   * @param pid Pid des Objekts, ber das gesendet werden soll.
   * @param atg Attributgruppe der zuu sendenden Daten.
   * @param asp Aspekt der zum Senden verwendet werden soll.
   * @param d Zu versendende Daten.
   * @throws DataNotSubscribedException Fehler bei der Subskription
   * @throws SendSubscriptionNotConfirmed Fehler bei der Subskription
   */

  public void sendData(final String pid, final String atg, final String asp, final Data d) 
      throws DataNotSubscribedException, SendSubscriptionNotConfirmed {
    DataDescription dd = new DataDescription(davConnection.getDataModel().getAttributeGroup(atg), 
        davConnection.getDataModel().getAspect(asp));
    ResultData data = new ResultData(davConnection.getDataModel().getObject(pid), dd, davConnection.getTime(), d);
    davConnection.sendData(data);
  }
  
  /**
   * Sende Daten an den Datenverteiler.
   * 
   * @param obj Objekt, ber das gesendet werden soll.
   * @param atg Attributgruppe der zuu sendenden Daten.
   * @param asp Aspekt der zum Senden verwendet werden soll.
   * @param d Zu versendende Daten.
   * @throws DataNotSubscribedException Fehler bei der Subskription
   * @throws SendSubscriptionNotConfirmed Fehler bei der Subskription
    */

  public void sendData(final SystemObject obj, final AttributeGroup atg, final Aspect asp, final Data d) 
      throws DataNotSubscribedException, SendSubscriptionNotConfirmed {
    DataDescription dd = new DataDescription(atg, asp);
    ResultData data = new ResultData(obj, dd, davConnection.getTime(), d);
    davConnection.sendData(data);
  }

  @Override
  public JAXBContext getJAXBContext() throws JAXBException {
    String con =  "eu.datex2.schema.dynamisch.messquerschnitte.verkehrslage";
    JAXBContext jbc = JAXBContext.newInstance(con);
      return jbc;
  }

  @Override
  public boolean initialisiere() {
    this.davConnection = plugin.getDaVConnection();
    logPrefix = "In Kontext '" + plugin.getKonfigurationsModul().getContext() + "': ";
    atg = davConnection.getDataModel().getAttributeGroup(ATG_MQ_ZUSTAND);
    aspPid = plugin.getKonfigurationsModul().getString("ImportAspekt");
    DataDescription dd = new DataDescription(davConnection.getDataModel().getAttributeGroup(ATG_MQ_ZUSTAND),
        davConnection.getDataModel().getAspect(aspPid));
    for (ObjekteAtgAsp x : plugin.getObjects()) {
      try {
        davConnection.subscribeSender(this, x.getSystemObjekte(), dd, SenderRole.source());
      } catch (OneSubscriptionPerSendData e) {
        LOGGER.error(logPrefix + "Kann keine Sendesteuerung fr " + dd + "erhalten.");
        return false;
      }
    }
    return true;
  }

  @Override
  public void dataRequest(final SystemObject arg0, final DataDescription arg1, final byte arg2) {
  }

  @Override
  public boolean isRequestSupported(final SystemObject arg0, final DataDescription arg1) {
    return false;
  }
}
